/*
 * mixfill.c - How to mix your nitrox/trimix using partial pressure method 
 * and topping with air.
 *
 * (c) 2002 Andreas Beck <becka@bedatec.de>
 * Thanks for helpful comments from Dirk Macke <Dirk.Macke@eed.ericsson.se>
 *
 * This program is free software; you can redistribute it and/or modify it      
 * under the terms of the GNU General Public License as published by the        
 * Free Software Foundation; either version 2 of the License, or (at your       
 * option) any later version.                                                   
 *                                                                              
 * This program is distributed in the hope that it will be useful, but          
 * WITHOUT ANY WARRANTY; without even the implied warranty of                   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General    
 * Public License for more details.                                             
 *                                                                              
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,      
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                        
 */

#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <string.h>
#include <limits.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "rcfile.h"
#include "convert.h"

#define O2PERCAIR 0.2095

struct parms {
	double initpress,inito2perc,initHeperc;
	double wantpress,wanto2perc,wantHeperc;
	int metric;
} parms = {
	  100000.0,O2PERCAIR,0.0,	/*   1 bar air remaining   */
	20000000.0,     0.32,0.0,	/* 200 bar EAN32 wanted    */
	1				/* metric output preferred */
};

struct air {
	double o2percent;
} air= {
	O2PERCAIR
};

int read_config(void) {

	char *path,*optarg;
	char buffer[1024];
	FILE *in;

	path=findrc(NULL,"mixfillrc",R_OK);
	if (!path) return 1;
	
	in=fopen(path,"r");
	if (!in) {
		fprintf(stderr,"Warning: Cannot fopen(%s), though findrc() said read o.k.\n",path);
		return 1;
	}
	while(fgets(buffer,sizeof(buffer),in)) {

		if ((optarg=strchr(buffer,'\n'))) 
			*optarg='\0';	/* Strip newline */
		if (*buffer=='#'||*buffer=='\0') 
			continue;	/* Strip comments*/

		if (NULL==(optarg=strchr(buffer,':'))) {	/* Find delimiter */
			fprintf(stderr,"Ignoring malformed line in config file %s:\n%s\n",path,buffer);
			continue;
		}
		*optarg++='\0';/* stop */
		while(isspace(*optarg)) optarg++;
		/*  */ if (0==strcasecmp(buffer,"InitialPressure")) {
			parms.initpress=atof(optarg);
		} else if (0==strcasecmp(buffer,"InitialO2Percent")) {
			parms.inito2perc=atof(optarg)/100.0;
		} else if (0==strcasecmp(buffer,"InitialHePercent")) {
			parms.initHeperc=atof(optarg)/100.0;
		} else if (0==strcasecmp(buffer,"WantedPressure")) {
			parms.wantpress=atof(optarg);
		} else if (0==strcasecmp(buffer,"WantedO2Percent")) {
			parms.wanto2perc=atof(optarg)/100.0;
		} else if (0==strcasecmp(buffer,"WantedHePercent")) {
			parms.wantHeperc=atof(optarg)/100.0;
		} else if (0==strcasecmp(buffer,"Units")) {
			parms.metric=(*optarg=='m'||*optarg=='M');
		} else {
			fprintf(stderr,"Ignoring unknown option in %s:\n%s\n",path,buffer);
		}
	}
	
	fclose(in);
	return 0;
}

double tocurrpressunit(double si) {
	if (parms.metric) {
		return convert(DIM_PRESSURE,UNIT_PASCAL,UNIT_BAR,si,NULL);
	} else {
		return convert(DIM_PRESSURE,UNIT_PASCAL,UNIT_PSI,si,NULL);
	}
}

char *currpressunit() {
	if (parms.metric) {
		return "bar";
	} else {
		return "PSI";
	}
}

int main(int argc, char **argv) {
  
	double needO2,needN2,needHe,needAir;
	
	read_config();

	while(1) {
	
		static struct option long_options[] = {
			{ "pressurein" 	,1,0,'p' },
			{ "o2in"       	,1,0,'o' },
			{ "hein"	,1,0,'h' },
			{ "pressureout"	,1,0,'P' },
			{ "o2out"	,1,0,'O' },
			{ "heout"	,1,0,'H' },
			{ "imperial"	,1,0,'i' },
			{ "metric"	,1,0,'m' },
			{ 0,0,0,0 }
		};
		int c;
		
		c=getopt_long(argc,argv,"p:P:o:O:h:H:im",long_options,NULL);
		if (c==-1) break;
		
		switch(c) {
		
			case 'p': 
				parms.initpress=atof(optarg);
				if (parms.metric) {
					parms.initpress=convert(DIM_PRESSURE,UNIT_BAR,UNIT_PASCAL,parms.initpress,NULL);
				} else {
					parms.initpress=convert(DIM_PRESSURE,UNIT_PSI,UNIT_PASCAL,parms.initpress,NULL);
				}
				break;
			case 'o':
				parms.inito2perc=atof(optarg)/100.0;
				break;
			case 'h':
				parms.initHeperc=atof(optarg)/100.0;
				break;
			case 'P':
				parms.wantpress=atof(optarg);
				if (parms.metric) {
					parms.wantpress=convert(DIM_PRESSURE,UNIT_BAR,UNIT_PASCAL,parms.wantpress,NULL);
				} else {
					parms.wantpress=convert(DIM_PRESSURE,UNIT_PSI,UNIT_PASCAL,parms.wantpress,NULL);
				}
				break;
			case 'O':
				parms.wanto2perc=atof(optarg)/100.0;
				break;
			case 'H':
				parms.wantHeperc=atof(optarg)/100.0;
				break;
			case 'm':
				parms.metric=1;
				break;
			case 'i':
				parms.metric=0;
				break;
			default :
				fprintf(stderr, "Usage: %s [-m/i] [-p initpressure] [-o initO2perc] [-h initHeperc]\n"
						"                 [-P wantpressure] [-O wantO2perc] [-H wantHeperc]\n",argv[0]);
				return 1;
		}
	}	

	printf("\nNitrox/Trimix filling calculation:\n\n");
	printf("Have %6.1f %s with %5.1f%% O_2 and %5.1f%% He.\n",
		tocurrpressunit(parms.initpress),currpressunit(),parms.inito2perc*100.0,parms.initHeperc*100.0);
	printf("Want %6.1f %s with %5.1f%% O_2 and %5.1f%% He.\n",tocurrpressunit(parms.wantpress),currpressunit(),parms.wanto2perc*100.0,parms.wantHeperc*100.0);

	needO2=     parms.wanto2perc *parms.wantpress-     parms.inito2perc *parms.initpress;
	needHe=     parms.wantHeperc *parms.wantpress-     parms.initHeperc *parms.initpress;
	needN2=(1.0-parms.wanto2perc-parms.wantHeperc)*parms.wantpress
	      -(1.0-parms.inito2perc-parms.initHeperc)*parms.initpress;
	printf("\nNeeded partial pressures: %6.1f %s O_2, %6.1f %s N_2, %6.1f %s He\n",
		tocurrpressunit(needO2),currpressunit(),
		tocurrpressunit(needN2),currpressunit(),
		tocurrpressunit(needHe),currpressunit());

	needAir=needN2/(1.0-air.o2percent);
	printf("\n=>Needed Air pressure: %.1f %s\n",
		tocurrpressunit(needAir),currpressunit());

	needO2-=needAir*air.o2percent;
	needN2-=needAir*(1-air.o2percent);

	printf("=>Remaining O_2 pressure needed: %.1f %s\n",
		tocurrpressunit(needO2),currpressunit());

	if (needAir<-0.1||needO2<-0.1||needN2<-0.1||needN2>0.1||needHe<-0.1) {
		printf( "\nSorry, the desired mix cannot be reached by partial pressure filling\n"
			"with air and pure Oxygen. Please check your values.\n"
			"If they are ok, you may have to drain the cylinder a bit first.\n");
		return 1;
	}

	printf("\nFilling plan:\n\n");
	printf("Start with %6.1f %s (%.1f%% O_2, %.1f%% He)\n",
		tocurrpressunit(parms.initpress),currpressunit(),
		parms.inito2perc*100.0,parms.initHeperc*100.0);
	if (needO2>0.05)
		printf("Fill up to %6.1f %s with 100%% O_2\n",
			tocurrpressunit(parms.initpress+needO2),currpressunit());
	if (needHe>0.05)
		printf("Fill up to %6.1f %s with 100%% He\n",
			tocurrpressunit(parms.initpress+needO2+needHe),currpressunit());
	printf("Top off to %6.1f %s with Air\n",
		tocurrpressunit(parms.initpress+needO2+needHe+needAir),currpressunit());

	return(0);
}

