/*
 * vyper2udcf.c Converts downloaded profile data from Suunto Vyper and 
 * compatible computers to UDCF.
 *
 * (c) 2002 Andreas Beck <becka@bedatec.de>
 *
 * Lots of good suggestions and probably also some code by
 * (c) 2002 Mike Brodbelt (mike@coruscant.demon.co.uk)
 *
 * This program is free software; you can redistribute it and/or modify it      
 * under the terms of the GNU General Public License as published by the        
 * Free Software Foundation; either version 2 of the License, or (at your       
 * option) any later version.                                                   
 *                                                                              
 * This program is distributed in the hope that it will be useful, but          
 * WITHOUT ANY WARRANTY; without even the implied warranty of                   
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General    
 * Public License for more details.                                             
 *                                                                              
 * You should have received a copy of the GNU General Public License along      
 * with this program; if not, write to the Free Software Foundation, Inc.,      
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                        
 */

#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <ctype.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "convert.h"
#include "rcfile.h"

char *fgets_nocomment(char *s,int size,FILE *stream) {

	char *rc;
	
	do {
		rc=fgets(s,size,stream);
	} while(rc&&*rc=='#');	/* Ignore comment lines. */
	return rc;
}


struct configdata {
	double  dctimezone;
	double  divetimezone;
} config = {
	0,		/* DC is set to UTC */
	0		/* Dive was in a place using UTC */
};

int gethex(char *src,unsigned char *dst,int *maxlen) {

	int len=0;
	int hlp;

	while(*src&&*src!='"') src++;
	if (*src) src++;else return 1;	/* Found no starting quote */

	while(len<=*maxlen) {
		switch(*src) {
			case '0':case '1':case '2':case '3':case '4':
			case '5':case '6':case '7':case '8':case '9':
				hlp=*src-'0';break;
			case 'a':case 'b':case 'c':case 'd':case 'e':case 'f':
				hlp=*src-'a'+10;break;
			case 'A':case 'B':case 'C':case 'D':case 'E':case 'F':
				hlp=*src-'A'+10;break;
			case '"': /* Done */
				*maxlen=len; return 0;
			default:
				fprintf(stderr,"Error parsing hex digit %c.\n",*src);
				return 1;
		}
		hlp<<=4;src++;
		switch(*src) {
			case '0':case '1':case '2':case '3':case '4':
			case '5':case '6':case '7':case '8':case '9':
				hlp|=*src-'0';break;
			case 'a':case 'b':case 'c':case 'd':case 'e':case 'f':
				hlp|=*src-'a'+10;break;
			case 'A':case 'B':case 'C':case 'D':case 'E':case 'F':
				hlp|=*src-'A'+10;break;
			default:
				fprintf(stderr,"Error parsing hex digit %c.\n",*src);
				return 1;
		}
		*dst++=hlp;src++;len++;
	}
	return 1;	/* maxlen overflowed. */
}


void print_preamble(void) {
	printf("<PROFILE UDCF=\"1\">\n");
	printf("<!-- generated by vyper2udcf (c) 2002 Andreas Beck [becka@bedatec.de] -->\n");
	printf("<UNITS>si</UNITS>\n");
}

void print_postamble(void) {
	printf("<!-- generated by vyper2udcf (c) 2002 Andreas Beck [becka@bedatec.de] -->\n");
	printf("</PROFILE>\n");
}

static enum MODEL { M_UNKNOWN, M_VYPER, M_SPYDER_OLD, M_SPYDER_NEW } cmodel=M_UNKNOWN;

static char serialnum[32]="none";

int print_dc_data(unsigned char *vypermem,int memlen) {

	char *model="unknown";

	switch(vypermem[0x24]) {
		case 20:
		case 30:
		case 40:fprintf(stderr,"Probably Spyder. Checking further.\n");
			if (vypermem[0x16]==0x01&&vypermem[0x17]==0x01) {
				cmodel=M_SPYDER_OLD;
				model="old Spyder";
				goto showmodel;
			} else if (vypermem[0x16]==0x01&&vypermem[0x17]==0x02) {
				cmodel=M_SPYDER_NEW;
				model="new Spyder";
				goto showmodel;
			} else {
				fprintf(stderr,"%s:%d:Cannot identify Spyder. Firmware:%02x.%02x. Sorry.\n",__FILE__,__LINE__,vypermem[0x16],vypermem[0x17]);
				return 1;
			}
			break;
		case 0x0c:
			model="old Cobra/Vyper";
			cmodel=M_VYPER;
			showmodel:
			fprintf(stderr,"DC dump is from %s.\n",model);
			break;
		case 0x0a:
			model="new Vyper";goto showmodel;
		case 0x03:
			model="Stinger";goto showmodel;
		case 0x04:
			model="Mosquito";goto showmodel;
		default:
			fprintf(stderr,"Unknown DC %02x detected. Aborting.\n",vypermem[0x24]);
			return 1;
	}
	printf( "<DEVICE>\n"
		" <VENDOR>Suunto</VENDOR>\n"
		" <MODEL>%s</MODEL>\n"
		" <VERSION>1</VERSION>\n"
		"</DEVICE>\n",model);

	/* This works generically on both DCs */
	printf("<ADDINFO>\n");
	{
		char buffer[31];
		memcpy(buffer,vypermem+0x2c,30);
		buffer[30]=0;
		printf(" <PERSINFO>%s</PERSINFO>\n",buffer);
	}

	/* 1e/1f holds maximum depth reached ever. */
	printf(" <!-- MAXDEPTH: %.2f m -->\n",
		convert(DIM_LENGTH,UNIT_FEET,UNIT_METER,(double)((vypermem[0x1e]<<8)+vypermem[0x1f])/128.0,NULL));

	/* 20/21 holds total dive time in minutes. */
	printf(" <!-- TOTALDIVETIME: %f seconds -->\n",
		convert(DIM_TIME,UNIT_MINUTE,UNIT_SECOND,(double)((vypermem[0x20]<<8)+vypermem[0x21]),NULL));

	/* 22/23 holds total number of dives. */
	printf(" <TOTALDIVES>%d</TOTALDIVES>\n",(vypermem[0x22]<<8)+vypermem[0x23]);

	switch(cmodel) {
		case M_SPYDER_OLD:
			/* 18/19/1a/1b has the serial number. */
			printf(" <SERIALID>%2d%2d%5d</SERIALID>\n",vypermem[0x18],vypermem[0x19],((vypermem[0x1a]<<8)+vypermem[0x1b]));
			sprintf(serialnum,"%2d%2d%5d",vypermem[0x18],vypermem[0x19],((vypermem[0x1a]<<8)+vypermem[0x1b]));
			goto spyder_misc;
		case M_SPYDER_NEW:
			/* 18/19/1a/1b has the serial number. */
			printf(" <SERIALID>%d%5d</SERIALID>\n",((vypermem[0x18]<<8)+vypermem[0x19]),((vypermem[0x1a]<<8)+vypermem[0x1b]));
			sprintf(serialnum,"%d%5d",((vypermem[0x18]<<8)+vypermem[0x19]),((vypermem[0x1a]<<8)+vypermem[0x1b]));
			spyder_misc:
			/* 24/25 is interval/settings for spyder. */
			printf("  <!-- Currently set interval on Spyder is %d s -->\n",vypermem[0x24]);
			printf("  <!-- Currently set alt/personal on Spyder is %dA%dP -->\n",vypermem[0x25]%3,vypermem[0x25]/3);
			goto sanitize_serial;
		default:
			/* 24/25 is type information for others. */
			/* 26/27/28/29 has the serial number. */
			printf(" <SERIALID>%2d%2d%2d</SERIALID>\n",((vypermem[0x26]<<8)+vypermem[0x27]),vypermem[0x28],vypermem[0x29]);
			sprintf(serialnum,"%2d%2d%2d",((vypermem[0x26]<<8)+vypermem[0x27]),vypermem[0x28],vypermem[0x29]);
			sanitize_serial:
			{ char *hlp=serialnum; for(;*hlp;hlp++) if (*hlp==' ') *hlp='_'; }
			break;
	}
	printf("</ADDINFO>\n");
	return 0;
}

int print_dive_data(unsigned char *rc,int len) {

	double depth=0.0;
	int start,stop,i;

	struct tm compreadtime,utctime,spottime;
	time_t unixtime,unixspottime;
		
	const char* wdaytab[]={"Sun","Mon","Tue","Wed","Thu","Fri","Sat","Invalid"};

	if (len<14) {
		fprintf(stderr,"Strange short dive (len=%d) - ignoring.\n",len);
		return 1;
	}

	printf("<REPGROUP>\n");
	printf(" <DIVE>\n");
	printf("  <!-- TC-Divenum: %d -->\n",rc[2]);

	/* Now work out the time stuff - really scary.
	 */
	switch(cmodel) {
		case M_SPYDER_NEW:
		case M_SPYDER_OLD:
			compreadtime.tm_sec =0;
			compreadtime.tm_min =rc[10];
			compreadtime.tm_hour=rc[9];
			compreadtime.tm_mday=rc[8];
			compreadtime.tm_mon =rc[7]-1;
			compreadtime.tm_year= rc[6]>90 ? rc[6] : 100+rc[6];
			compreadtime.tm_wday=0; /* will recalc that. */
			compreadtime.tm_yday=0; /* will recalc that. */
			compreadtime.tm_isdst=-1;	/* don't know. */
			break;
		default:
			compreadtime.tm_sec =0;
			compreadtime.tm_min =rc[13];
			compreadtime.tm_hour=rc[12];
			compreadtime.tm_mday=rc[11];
			compreadtime.tm_mon =rc[10]-1;
			compreadtime.tm_year= rc[9]>90 ? rc[9] : 100+rc[9];
			compreadtime.tm_wday=0; /* will recalc that. */
			compreadtime.tm_yday=0; /* will recalc that. */
			compreadtime.tm_isdst=-1;	/* don't know. */
			break;
	}
	setenv("TZ","UTC",1);	/* force UTC - uh is that silly. */
	unixtime=mktime(&compreadtime);
	utctime =*gmtime(&unixtime);
	if (utctime.tm_hour!=compreadtime.tm_hour||
	    utctime.tm_min !=compreadtime.tm_min) {
		fprintf(stderr,"Warning: mktime() is not performing as expected. Times might be wrong.\n");
	}
	unixtime   -=config.dctimezone*60*60;
	utctime     =*gmtime(&unixtime);
	unixspottime=unixtime+config.divetimezone*60*60;
	spottime    =*gmtime(&unixspottime);

	printf("  <!-- Computer time reads %4d-%2d-%2d %2d:%02d:%02d (%s, Julian %3d) timezone is GMT%+4.1f -->\n",
		compreadtime.tm_year+1900,
		compreadtime.tm_mon +1,
		compreadtime.tm_mday,
		compreadtime.tm_hour,
		compreadtime.tm_min,
		compreadtime.tm_sec,
		wdaytab[compreadtime.tm_wday],
		compreadtime.tm_yday,
		config.dctimezone
		);
	printf("  <!-- UTC time is         %4d-%2d-%2d %2d:%02d:%02d (%s, Julian %3d) Unixtime %ld -->\n",
		utctime.tm_year+1900,
		utctime.tm_mon +1,
		utctime.tm_mday,
		utctime.tm_hour,
		utctime.tm_min,
		utctime.tm_sec,
		wdaytab[utctime.tm_wday],
		utctime.tm_yday,
		(long) unixtime
		);
	printf("  <!-- Divesite time is    %4d-%2d-%2d %2d:%02d:%02d (%s, Julian %3d) timezone is GMT%+4.1f -->\n",
		spottime.tm_year+1900,
		spottime.tm_mon +1,
		spottime.tm_mday,
		spottime.tm_hour,
		spottime.tm_min,
		spottime.tm_sec,
		wdaytab[spottime.tm_wday],
		spottime.tm_yday,
		config.divetimezone
		);
	printf("  <!-- parseable date/time information is given in UTC -->\n");
	printf("  <DATE><YEAR>%d</YEAR><MONTH>%d</MONTH><DAY>%d</DAY></DATE>\n",
		utctime.tm_year+1900,utctime.tm_mon+1,utctime.tm_mday);
	printf("  <TIME><HOUR>%d</HOUR><MINUTE>%d</MINUTE></TIME>\n",
		utctime.tm_hour,utctime.tm_min);

	if (rc[0]==0&&rc[1]==0)
		printf("  <SURFACEINTERVAL>infinity</SURFACEINTERVAL>\n");
	else
		printf("  <SURFACEINTERVAL>%f</SURFACEINTERVAL> <!-- unit seconds -->\n",
			 convert(DIM_TIME,UNIT_MINUTE,UNIT_SECOND,(double)rc[0],NULL)
			+convert(DIM_TIME,UNIT_HOUR  ,UNIT_SECOND,(double)rc[1],NULL));
			/*	was rc[0]+60*rc[1] minutes */

	switch(cmodel) {
		case M_SPYDER_NEW:
		case M_SPYDER_OLD:
			printf("  <TEMPERATURE>%f</TEMPERATURE> <!-- Kelvin -->\n",
				convert(DIM_TEMPERATURE,UNIT_CELSIUS,UNIT_KELVIN,(double)rc[len-2],NULL));
				/* rc[len-2]); */
			break;
		default:
			printf("  <TEMPERATURE>%f</TEMPERATURE> <!-- Kelvin -->\n",
				convert(DIM_TEMPERATURE,UNIT_CELSIUS,UNIT_KELVIN,(double)rc[len-4],NULL));
				/* rc[len-4]); */
			printf("  <!-- TEMPERATURE AIR at start %f -->\n",
				convert(DIM_TEMPERATURE,UNIT_CELSIUS,UNIT_KELVIN,(double)rc[8],NULL));
				/* rc[8] */
			printf("  <!-- TEMPERATURE at end of dive %f -->\n",
				convert(DIM_TEMPERATURE,UNIT_CELSIUS,UNIT_KELVIN,(double)rc[len-3],NULL));
				/* rc[len-3] */
			printf("  <!-- OLF at end of dive %d -->\n",
				rc[len-1]);
			break;
	}
	printf("  <DENSITY>1030.000</DENSITY> <!-- kg/m^3 -->\n");
	switch(cmodel) {
		case M_SPYDER_NEW:
		case M_SPYDER_OLD:
			switch(rc[5]%3) {
				case 0:
					printf("  <ALTITUDE>300</ALTITUDE>\n");break;
				case 1:
					printf("  <ALTITUDE>1500</ALTITUDE>\n");break;
				case 2:
					printf("  <ALTITUDE>3000</ALTITUDE>\n");break;
			}
			break;
		default:
			switch(rc[4]%3) {
				case 0:
					printf("  <ALTITUDE>300</ALTITUDE>\n");break;
				case 1:
					printf("  <ALTITUDE>1500</ALTITUDE>\n");break;
				case 2:
					printf("  <ALTITUDE>3000</ALTITUDE>\n");break;
			}
			break;
	}
	printf("  <GASES>\n");
	printf("   <MIX>\n");
	if (rc[6]==0) {	/* Air mode */
		printf("    <MIXNAME>Air</MIXNAME>\n");
	} else {
		printf("    <MIXNAME>EAN%d</MIXNAME>\n",rc[6]);
	}
	printf("    <TANK>\n");
	printf("     <TANKVOLUME>%f</TANKVOLUME>\n",
		convert(DIM_VOLUME,UNIT_LITER,UNIT_METER_3,(double)10.0,NULL));
	printf("     <PSTART>%f</PSTART>\n",
		convert(DIM_PRESSURE,UNIT_BAR,UNIT_PASCAL,(double)rc[5]*2,NULL));
		/* rc[5]*2 bar */
	printf("     <PEND>%f</PEND>\n",
		convert(DIM_PRESSURE,UNIT_BAR,UNIT_PASCAL,(double)rc[len-2]*2,NULL));
		/* rc[len-2]*2 bar */
	printf("    </TANK>\n");
	if (cmodel==M_SPYDER_NEW||cmodel==M_SPYDER_OLD||
	    rc[6]==0) {	/* Air mode */
		printf("    <O2>0.21</O2>\n");
		printf("    <N2>0.79</N2>\n");
		printf("    <HE>0.00</HE>\n");
	} else {
		printf("    <O2>0.%02d</O2>\n",rc[6]);
		printf("    <N2>0.%02d</N2>\n",100-rc[6]);
		printf("    <HE>0.00</HE>\n");
	}
	printf("   </MIX>\n");
	printf("  </GASES>\n");
	printf("  <DELTAMODE/>\n");
	printf("  <SAMPLES>\n");
	printf("   <DELTA>%f</DELTA>\n",(double)rc[3]);
	if (cmodel==M_SPYDER_NEW||cmodel==M_SPYDER_OLD||
	    rc[6]==0) {	/* Air mode */
		printf("   <SWITCH>Air</SWITCH>\n");
	} else {
		printf("  <SWITCH>EAN%d</SWITCH>\n",rc[6]);
	}

	/* Start at 0 depth. */
	printf("   <D>%.2f</D>\n",
		convert(DIM_LENGTH,UNIT_FEET,UNIT_METER,(double)depth,NULL));
	switch(cmodel) {
		case M_SPYDER_NEW:
		case M_SPYDER_OLD:
			start=11;
			stop=len-2;
			break;
		default:
			start=14;
			stop=len-4;
			break;
	}
	for(i=start;i<stop;i++) {
		switch(rc[i]) {
			case 0x7a:printf("   <ALARM>ASCENT</ALARM>\n");break;
			case 0x7b:printf("   <ALARM>ERROR</ALARM><!-- ATTENTION -->\n");break;
			case 0x7c:printf("   <ALARM>ERROR</ALARM><!-- BOOKMARK -->\n");break;
			case 0x7d:printf("   <ALARM>SURFACE</ALARM>\n");break;
			case 0x7e:printf("   <ALARM>DECO</ALARM>\n");break;
			case 0x7f:printf("   <ALARM>ERROR</ALARM><!-- CEILING VIOLATED -->\n");break;
			case 0x80:printf("   <!-- END OF DIVE -->\n");break;
			case 0x81:
				switch(cmodel) {
					case M_SPYDER_NEW:
					case M_SPYDER_OLD:
						printf("   <ALARM>ASCENT</ALARM><!-- Is that right? Please report if you see this! -->\n");break;
						break;
					default:
						printf("   <ALARM>ERROR</ALARM><!-- MANDATORY SAFETY STOP VIOLATED -->\n");break;
						break;
				}
				break;
			default: {
				double d=rc[i];
				if (d>127.5) d-=256.0;
				depth+=d;
			}
				printf("   <D>%.2f</D>\n",
					convert(DIM_LENGTH,UNIT_FEET,UNIT_METER,(double)depth,NULL));
		}
	}
	/* Stop at 0 depth. */
	depth=0.0;
	printf("   <D>%.2f</D>\n",
		convert(DIM_LENGTH,UNIT_FEET,UNIT_METER,(double)depth,NULL));
	printf("  </SAMPLES>\n");
	printf(" </DIVE>\n");
	printf("</REPGROUP>\n");

	return 0;
}

int read_config(char *suffix) {

	char *path,*optarg;
	char buffer[1024];
	FILE *in;

	strcpy(buffer,"vyper2udcfrc");
	if (suffix) {
		strcat(buffer,".");
		strcat(buffer,suffix);
	}

	path=findrc(NULL,buffer,R_OK);
	if (!path) return 1;
	
	in=fopen(path,"r");
	if (!in) {
		fprintf(stderr,"Warning: Cannot fopen(%s), though findrc() said read o.k.\n",path);
		return 1;
	}
	while(fgets(buffer,sizeof(buffer),in)) {

		if ((optarg=strchr(buffer,'\n'))) 
			*optarg='\0';	/* Strip newline */
		if (*buffer=='#'||*buffer=='\0') 
			continue;	/* Strip comments*/

		if (NULL==(optarg=strchr(buffer,':'))) {	/* Find delimiter */
			fprintf(stderr,"Ignoring malformed line in config file %s:\n%s\n",path,buffer);
			continue;
		}
		*optarg++='\0';/* stop */
		while(isspace(*optarg)) optarg++;
		/*  */ if (0==strcasecmp(buffer,"dctimezone")) {
			config.dctimezone=atoi(optarg);
		} else if (0==strcasecmp(buffer,"divetimezone")) {
			config.divetimezone=atoi(optarg);
		} else {
			fprintf(stderr,"Ignoring unknown option in %s:\n%s\n",path,buffer);
		}
	}
	
	fclose(in);
	return 0;
}


int main(int argc, char **argv) {
  
	unsigned char rc[16*1024];
	unsigned char vypermem[256];
	int len;
	int cmdline_override=0;

	read_config(NULL);
	
	while(1) {
	
		static struct option long_options[] = {
			{ "dctimezone",1,0,'Z' },
			{ "divetimezone",1,0,'z' },
			{ 0,0,0,0}
		};
		int c;
		
		c=getopt_long(argc,argv,"Z:z:",long_options,NULL);
		if (c==-1) break;
		
		switch(c) {
			case 'z':
				config.divetimezone=atof(optarg);
				cmdline_override=1;
				break;
			case 'Z':
				config.dctimezone=atof(optarg);
				cmdline_override=1;
				break;
			default :
				fprintf(stderr,"Usage: %s [-Z dc-timezone-offset] [-z divespot-timezone-offset]\n",argv[0]);
				return 1;
		}
	}	

	if (config.dctimezone  >23||config.dctimezone  <-23 ||
	    config.divetimezone>23||config.divetimezone<-23) {
		fprintf(stderr,"%s:Warning: Timezone data is way off +/-12 should do.\n",argv[0]);
	}

	memset(vypermem,0,sizeof(vypermem));	/* Erase virtual Vyper memory */

	print_preamble();
	while(1) {
		if (NULL==fgets_nocomment(rc,sizeof(rc),stdin)) {
			fprintf(stderr,"Unexpected EOF on input stream.\n");
			break;
		}
		/*  */ if (0==strncmp(rc,"VyperRead(",10)) {
			int base,len,maxlen;
			char *datastart;
			base=len=-1;
			sscanf(rc+10,"%i,%i",&base,&len);
			maxlen=len;
			if (base<0||len<0||base+len>=sizeof(vypermem)) {
				fprintf(stderr,"Huh - overflowing vypermem.\nAborting.\n");
				return 1;
			}
			datastart++;
			if (gethex(rc,vypermem+base,&maxlen) || maxlen!=len) {
				fprintf(stderr,"Problem getting data for VyperRead.\n");
				return 1;
			}
		} else if (0==strncmp(rc,"VyperFirstProfile=\"",19)) {
			print_dc_data(vypermem,sizeof(vypermem));
			if (!cmdline_override) {
				fprintf(stderr,"Serial-suffix is %s\n",serialnum);
				read_config(serialnum);
			}
			goto showprofile;
		} else if (0==strncmp(rc,"VyperProfile=\"",14)) {
			int maxlen;
			showprofile:
			maxlen=sizeof(rc);
			len=gethex(rc,rc,&maxlen);
			if (len) {
				fprintf(stderr,"Problem processing a profile.\n");
				return 1;
			}
			if (maxlen>0) print_dive_data(rc,maxlen);
			;
		} else if (0==strncmp(rc,"VyperEnd",8)) {
			break;
		} else {
			fprintf(stderr,"Cannot parse the follwong line:\n%s\nAborting.",rc);
			return 1;
		}
	}
	print_postamble();

	return(0);
}

